#!/usr/bin/env python
# -*- coding: utf-8 -*-
from math import sqrt
from roller_constant_for import Gradient as fg, Plan as fy
from roller_constant_key import Model as md, Option as ok
from roller_constant_fu import Fu
from roller_one import Hat
from roller_one_fu import Lay, Sel
from roller_render_gradient_light import GradientLight
from roller_render_hub import RenderHub
import colorsys
import gimpfu as fu

em = Fu.Emboss
pdb = fu.pdb
sn = Fu.SolidNoise
um = Fu.UnsharpMask
X_IS_1 = Y_IS_1 = 1


def do_grid(one, edge, span, half):
    """
    Create a list of selections for images in the grid.

    one: One
        Has variables.

    edge: int
        width of edge

    span: int
        width of frame less edge

    half: int
        half of the frame width

    return: list
        of selection rectangle
    """
    row, column = one.grid.division
    is_merge_cell = one.grid.is_merge_cell
    e = one.grid.grid_d
    rect = []

    if not is_merge_cell:
        s = 1

    # Load the image selections. Calculate the frame bounds.
    # Collect the frame bounds for plate clearing.
    # Do one image at a time:
    for r in range(row):
        for c in range(column):
            if is_merge_cell:
                s = e[ok.PER_CELL][r][c]

            # Is it a topleft cell?
            if s != (-1, -1):
                # Yes, it is a topleft cell:
                a = process_image(one, r, c, edge, span, half)
                if a:
                    rect.append(a)
    return rect


def do_rounded_edge(j, z, edge):
    """
    Do color blend effect using a color and its luminosity.

    The round effect is calculated using the unit circle formula.
    So 'x' is the luminosity step, and 'y' is the luminosity result.

        x**2 + y**2 = 1

    j: GIMP image
        work-in-progress

    z: layer
        to receive effect

    edge: int
        the number of gradient edge steps
    """
    # the x-vector steps:
    f = 1. / (edge - 1)

    # '255.' is the high-end of the range for luminosity:
    q = colorsys.hsv_to_rgb(0., 0., 255. * sqrt(1. - f**2))

    Sel.item(z)
    for i in range(edge):
        if Sel.is_sel(j):
            a = pdb.gimp_selection_save(j)
            q = tuple([int(b) for b in q])

            Sel.grow(j, 1, 0)
            Sel.load(j, a, option=fu.CHANNEL_OP_SUBTRACT)
            Sel.fill(z, q)
            Sel.load(j, a, option=fu.CHANNEL_OP_ADD)

            # 'luminosity' is on the  y-vector of the unit circle:
            y = f * (i + 1)

            if y < 1:
                q = colorsys.hsv_to_rgb(0., 0., 255. * sqrt(1 - y**2))
            pdb.gimp_image_remove_channel(j, a)


def process_image(one, r, c, edge, span, half):
    """
    Calculate a selection rectangle for an image.

    one: One
        Has variables.

    r, c: int
        cell index

    edge: int
        width of edge

    span: int
        width of frame less edge

    half: int
        half of the frame width

    return: tuple or None
        a rectangle
        x, y, w, h
    """
    cat = Hat.cat
    j = cat.render.image
    cat.join_selection(one.model_name, r, c)
    if Sel.is_sel(j):
        # Calculate the frame bounds:
        Sel.grow(j, edge, 1)
        _, x, y, x1, y1 = pdb.gimp_selection_bounds(j)
        w = x1 - x + span
        h = y1 - y + span
        x -= half
        y -= half
        return x, y, w, h


class CutoutPlate:
    """Create a rectangular frame with cutout space for image material."""

    @staticmethod
    def do(one):
        """
        Is an image-effect template function.

        one: One
            Has variables.

        one: One
            Has variables.

        Return: layer
            Has Cutout Plate
        """
        cat = Hat.cat
        parent = one.parent
        d = one.d
        j = cat.render.image
        group = Lay.group(j, Lay.make_name(parent, one.k), parent=parent)
        w = d[ok.FRAME_WIDTH]
        half = w // 2
        edge = max(min(d[ok.BEVEL_EDGE_WIDTH], half), 1)
        half = max(half - edge, 1)
        span = max(w - edge * 2, 2)
        frame_layer = Lay.add(j, one.k, parent=group)
        rect = []

        Lay.color_fill(frame_layer, (255, 255, 255))

        # Cut out the image.
        # Use 'z' to set the alpha of the selection:
        Sel.make_layer_sel(one.image_layer)
        Sel.grow(j, edge, 1)
        Lay.clear_sel(frame_layer)

        if one.context == md.TABLE:
            rect = do_grid(one, edge, span, half)

        else:
            r = fy.CUSTOM_CELL
            a = process_image(one, r, r, edge, span, half)
            if a:
                rect = [a]

        pdb.gimp_selection_none(j)

        # Load the expanded rectangle selections:
        for i in rect:
            Sel.rect(j, *i)

        # Clear the material outside of the rectangle selections:
        Sel.invert_clear(frame_layer)

        do_rounded_edge(j, frame_layer, edge)

        # Begin frame decoration.
        # 'z', pattern:
        z = pattern_layer = Lay.add(
            j,
            one.k,
            parent=group,
            offset=Lay.offset(frame_layer)
        )
        frame_layer.mode = fu.LAYER_MODE_MULTIPLY

        Sel.item(z)
        RenderHub.set_fill_context(fg.FILL_DICT)
        pdb.gimp_context_set_pattern(d[ok.PATTERN])
        pdb.gimp_drawable_edit_bucket_fill(
            z,
            fu.FILL_PATTERN,
            X_IS_1, Y_IS_1
        )
        Sel.item(frame_layer)
        Sel.invert_clear(z)
        RenderHub.adjust_mean_value(z)

        if d[ok.NOISE_OPACITY]:
            # noise:
            z = Lay.add(j, one.k, parent=group)
            z.opacity = d[ok.NOISE_OPACITY]

            pdb.plug_in_solid_noise(
                j, z,
                sn.NO_TILEABLE,
                sn.YES_TURBULENT,
                d[ok.RANDOM_SEED],
                d[ok.DETAIL_LEVEL],
                sn.HORIZONTAL_SIZE_2,
                sn.VERTICAL_SIZE_2
            )
            pdb.plug_in_unsharp_mask(
                j, z,
                um.RADIUS_3,
                um.AMOUNT_54,
                um.THRESHOLD_0
            )

            # in case of transparency in the pattern:
            Sel.item(pattern_layer)

            sel = Hat.cat.save_short_term_sel()
            if sel:
                Sel.item(frame_layer)
                Sel.invert_clear(z)
                pdb.plug_in_colortoalpha(j, z, (127, 127, 127))
                Sel.item(z)
                pdb.gimp_selection_feather(j, d[ok.SOFTNESS])
                Sel.invert(j)
                Lay.clear_sel(z)
                Sel.load(j, sel)
                Sel.invert_clear(z)

        z = Lay.clone(frame_layer)
        z.mode = fu.LAYER_MODE_NORMAL
        frame_layer.opacity = 94.

        pdb.gimp_image_raise_layer_to_top(j, frame_layer)

        z = Lay.merge_group(group)

        z = GradientLight.apply_light(z, ok.OTHER_FRAME)
        one.shadow_layer = [z, one.image_layer]
        return z
